#include "stdafx.h"

#include <fstream>
#include <algorithm>

#include "Util.h"
#include "Logger.h"
#include "ChkWordList.h"

////////////////////////////////////////////////////////////////////////
// CHK [h
////////////////////////////////////////////////////////////////////////

//----------------------------------------------------------------------
// RXgN^
//----------------------------------------------------------------------

ChkWord::ChkWord() :
	m_wordType(ChkwAll),
	m_bRegex(false),
	m_bCase(false)
{
}

//----------------------------------------------------------------------
// RXgN^
//----------------------------------------------------------------------

ChkWord::ChkWord(const char* chkword, const WordType wdType, const bool bRegex, const bool bCase) :
	m_word(chkword),
	m_wordType(wdType),
	m_bRegex(bRegex),
	m_bCase(bCase)
{
}

//----------------------------------------------------------------------
// }b`邩
//----------------------------------------------------------------------

bool ChkWord::IsMatch(const char* subject, const char* fromname)
{
	bool bError;
	std::string errorWhat;

	if (m_wordType == ChkwSubject) {
		return Util::IsMatch(subject, m_word.c_str(), m_bRegex, !m_bCase, bError, errorWhat);
	} else if (m_wordType == ChkwFromname) {
		return Util::IsMatch(fromname, m_word.c_str(), m_bRegex, !m_bCase, bError, errorWhat);
	} else /*if (m_wordType == ChkwAll)*/ {
		bool matchSubject = Util::IsMatch(subject, m_word.c_str(), m_bRegex, !m_bCase, bError, errorWhat);
		if (matchSubject) {
			return true;
		}
		return Util::IsMatch(fromname, m_word.c_str(), m_bRegex, !m_bCase, bError, errorWhat);
	}
}

//----------------------------------------------------------------------
// CHK [h
//----------------------------------------------------------------------

std::string& ChkWord::GetWord()
{
	return m_word;
}

//----------------------------------------------------------------------
//  CHK [h̎
//----------------------------------------------------------------------

ChkWord::WordType ChkWord::GetWordType()
{
	return m_wordType;
}

//----------------------------------------------------------------------
// K\
//----------------------------------------------------------------------

bool ChkWord::IsRegex()
{
	return m_bRegex;
}

//----------------------------------------------------------------------
// p召
//----------------------------------------------------------------------

bool ChkWord::IsCase()
{
	return m_bCase;
}

////////////////////////////////////////////////////////////////////////
// CHK [hꗗ
////////////////////////////////////////////////////////////////////////

//----------------------------------------------------------------------
// RXgN^
//----------------------------------------------------------------------

ChkWordList::ChkWordList() :
	m_pcsChkWordList(0),
	m_sortKey(SortChkwType),
	m_bAscending(true)
{
}

//----------------------------------------------------------------------
// CHK [hɃ}b`̂邩𓾂
//----------------------------------------------------------------------

bool ChkWordList::IsMatch(const char* subject, const char* fromname)
{
	CSLock cslock(m_pcsChkWordList);

	for (std::map<std::string, ChkWord>::iterator it = m_chkWordList.begin(); it != m_chkWordList.end(); ++it) {
		if (it->second.IsMatch(subject, fromname)) {
			return true;
		}
	}

	return false;
}

//----------------------------------------------------------------------
// CHK [h݂邩𓾂
//----------------------------------------------------------------------

bool ChkWordList::IsExist(const char* chkword)
{
	CSLock cslock(m_pcsChkWordList);

	std::map<std::string, ChkWord>::iterator it = m_chkWordList.find(chkword);
	if (it != m_chkWordList.end()) {
		return true;
	}

	return false;
}

//----------------------------------------------------------------------
// CHK [h̏𓾂
//----------------------------------------------------------------------

std::map<std::string, ChkWord>& ChkWordList::GetChkWordList()
{
	CSLock cslock(m_pcsChkWordList);

	return m_chkWordList;
}

//----------------------------------------------------------------------
// CHK [ȟ𓾂
//----------------------------------------------------------------------

int ChkWordList::Count()
{
	CSLock cslock(m_pcsChkWordList);

	return m_chkWordList.size();
}

//----------------------------------------------------------------------
// CHK [hǉ
//----------------------------------------------------------------------

void ChkWordList::Add(const char* chkword, const ChkWord::WordType wdType, const bool bRegex, const bool bCase)
{
	CSLock cslock(m_pcsChkWordList);

	ChkWord chkw(chkword, wdType, bRegex, bCase);
	m_chkWordList[chkword] = chkw;

	// \[gp
	std::vector<std::string>::iterator it = std::find(m_sortedChkWord.begin(), m_sortedChkWord.end(), chkword);
	if (it == m_sortedChkWord.end()) {
		m_sortedChkWord.push_back(chkword);
	}
}

//----------------------------------------------------------------------
// 폜
//----------------------------------------------------------------------

void ChkWordList::Delete(const char* chkword)
{
#define __FUNCNAME__ "ChkWordList::Delete()"

	CSLock cslock(m_pcsChkWordList);

	m_chkWordList.erase(chkword);

	// \[gp
	std::vector<std::string>::iterator it = std::find(m_sortedChkWord.begin(), m_sortedChkWord.end(), chkword);
	if (it != m_sortedChkWord.end()) {
		m_sortedChkWord.erase(it);
	}

#undef __FUNCNAME__
}

//----------------------------------------------------------------------
// XV
//----------------------------------------------------------------------

void ChkWordList::Update(const char* chkword, const ChkWord::WordType wdType, const bool bRegex, const bool bCase)
{
	// std::map Ȃ̂ŁAgAꉞ֐𕪂Ă
	Add(chkword, wdType, bRegex, bCase);
}

//----------------------------------------------------------------------
// ǉ܂͍XV
//----------------------------------------------------------------------

void ChkWordList::AddOrUpdate(const char* chkword, const ChkWord::WordType wdType, const bool bRegex, const bool bCase)
{
	// std::map Ȃ̂ŁAgAꉞ֐𕪂Ă
	Add(chkword, wdType, bRegex, bCase);
}

//----------------------------------------------------------------------
// SNA
//----------------------------------------------------------------------

void ChkWordList::ClearAll()
{
	CSLock cslock(m_pcsChkWordList);

	m_chkWordList.clear();

	// \[gp
	m_sortedChkWord.clear();
}

//----------------------------------------------------------------------
// C|[g
//----------------------------------------------------------------------

void ChkWordList::Import(const char* filepath)
{
	CSLock cslock(m_pcsChkWordList);

	const char* separator = "\t";

	m_chkWordList.clear();
	m_sortedChkWord.clear();	// \[gp

	std::ifstream chkwordlistfile;
	chkwordlistfile.open(filepath, std::ios::in);
	if (chkwordlistfile) {
		std::string line;
		while (std::getline(chkwordlistfile, line)) {

			std::string chkwdata = Util::Trim(line);

			std::string chkword = "";
			ChkWord::WordType type = ChkWord::ChkwAll;
			bool bRegex = false;
			bool bCase = false;

			std::vector<std::string> chkwItems = Util::Split(chkwdata, *separator);
			if (0 < chkwItems.size()) {
				chkword = Util::Trim(chkwItems.at(0).c_str());
				chkword = chkword.substr(0, min(NUM_CHK_WORD_MAX, chkword.length()));
				chkword = Util::Trim(chkword);
			}
			if (1 < chkwItems.size()) {
				std::string stype = Util::Trim(chkwItems.at(1).c_str());
				if (stype == "All") {
					type = ChkWord::ChkwAll;
				} else if (stype == "Subject") {
					type = ChkWord::ChkwSubject;
				} else if (stype == "Fromname") {
					type = ChkWord::ChkwFromname;
				}
			}
			if (2 < chkwItems.size()) {
				std::string sRegex = Util::Trim(chkwItems.at(2).c_str());
				if (sRegex == "Regex") {
					bRegex = true;
				} else if (sRegex == "Text") {
					bRegex = false;
				}
			}
			if (3 < chkwItems.size()) {
				std::string sCase = Util::Trim(chkwItems.at(3).c_str());
				if (sCase == "Case") {
					bCase = true;
				} else if (sCase == "IgnoreCase") {
					bCase = false;
				}
			}

			if (0 < chkword.length()) {
				Add(chkword.c_str(), type, bRegex, bCase);
			}
		}

		chkwordlistfile.close();
	}
}

//----------------------------------------------------------------------
// GNX|[g
//----------------------------------------------------------------------

void ChkWordList::Export(const char* filepath)
{
	CSLock cslock(m_pcsChkWordList);

	std::ofstream chkwordlistfile;
	chkwordlistfile.open(filepath, std::ios::out | std::ios::trunc);
	if (chkwordlistfile) {

		for(std::map<std::string, ChkWord>::iterator it = m_chkWordList.begin(); it != m_chkWordList.end(); ++it) {

			char* wordTypeStr = "";
			switch (it->second.GetWordType()) {
			case ChkWord::ChkwAll:
			default:
				wordTypeStr = "All";
				break;
			case ChkWord::ChkwSubject:
				wordTypeStr = "Subject";
				break;
			case ChkWord::ChkwFromname:
				wordTypeStr = "Fromname";
				break;
			}

			chkwordlistfile << it->second.GetWord() << '\t'
				<< wordTypeStr << '\t'
				<< ((it->second.IsRegex()) ? "Regex" : "Text") << '\t'
				<< ((it->second.IsCase()) ? "Case" : "IgnoreCase")
				<< std::endl;
		}

		chkwordlistfile.close();
	}
}

//----------------------------------------------------------------------
// rp̃IuWFNgݒ肷
//----------------------------------------------------------------------

void ChkWordList::SetLockObject(CRITICAL_SECTION* pcsChkWordList)
{
	m_pcsChkWordList = pcsChkWordList;
}

//----------------------------------------------------------------------
// \[gs
//----------------------------------------------------------------------

//
// \[gp
//
class SortPredChkWord {
public:
	// RXgN^
	SortPredChkWord(ChkWordList::SortKey sortKey, bool bAscending, std::map<std::string, ChkWord>& chkWordList) :
		m_chkWordList(chkWordList)
	{
		m_sortKey = sortKey;
		m_bAscending = bAscending;
	}

	// r֐
	bool operator()(std::string& left, std::string& right) {
		bool result = false;

		std::map<std::string, ChkWord>::iterator itLeft = m_chkWordList.find(left);
		std::map<std::string, ChkWord>::iterator itRight = m_chkWordList.find(right);

		// Ce[^ end() ̏ꍇ̃`FbN͔Ô߂̃R[hŁA݂̎gł͋NȂ(͂)B
		if (itLeft == m_chkWordList.end() && itRight == m_chkWordList.end()) {
			result = true;
		} else if (itLeft == m_chkWordList.end() && itRight != m_chkWordList.end()) {
			result = true;
		} else if (itLeft != m_chkWordList.end() && itRight == m_chkWordList.end()) {
			result = false;
		} else {

			const int sortKeyNum = 4;
			ChkWordList::SortKey sortKeyArr[sortKeyNum];

			switch (m_sortKey) {
			case ChkWordList::SortChkwType:
			default:
				sortKeyArr[0] = ChkWordList::SortChkwType;
				sortKeyArr[1] = ChkWordList::SortChkwWord;
				sortKeyArr[2] = ChkWordList::SortChkwRegex;
				sortKeyArr[3] = ChkWordList::SortChkwCase;
				break;
			case ChkWordList::SortChkwWord:
				sortKeyArr[0] = ChkWordList::SortChkwWord;
				sortKeyArr[1] = ChkWordList::SortChkwType;
				sortKeyArr[2] = ChkWordList::SortChkwRegex;
				sortKeyArr[3] = ChkWordList::SortChkwCase;
				break;
			case ChkWordList::SortChkwRegex:
				sortKeyArr[0] = ChkWordList::SortChkwRegex;
				sortKeyArr[1] = ChkWordList::SortChkwWord;
				sortKeyArr[2] = ChkWordList::SortChkwType;
				sortKeyArr[3] = ChkWordList::SortChkwCase;
				break;
			case ChkWordList::SortChkwCase:
				sortKeyArr[0] = ChkWordList::SortChkwCase;
				sortKeyArr[1] = ChkWordList::SortChkwWord;
				sortKeyArr[2] = ChkWordList::SortChkwType;
				sortKeyArr[3] = ChkWordList::SortChkwRegex;
				break;
			}

			int i = 0;
			for ( ; i < sortKeyNum; ++i) {

				switch (sortKeyArr[i]) {
				case ChkWordList::SortChkwType:
				default:
					if (itLeft->second.GetWordType() < itRight->second.GetWordType()) {
						result = true;
					} else if (itLeft->second.GetWordType() == itRight->second.GetWordType()) {
						continue;
					} else {
						result = false;
					}
					break;

				case ChkWordList::SortChkwWord:
					{
						int comprc = itLeft->first.compare(itRight->first);
						if (comprc < 0) {
							result = true;
						} else if (comprc == 0) {
							continue;	// 蓾Ȃ(͂)BÔ
						} else {
							result = false;
						}
					}
					break;

				case ChkWordList::SortChkwRegex:
					if (!itLeft->second.IsRegex() && itRight->second.IsRegex()) {
						result = true;
					} else if (itLeft->second.IsRegex() == itRight->second.IsRegex()) {
						continue;
					} else {
						result = false;
					}
					break;

				case ChkWordList::SortChkwCase:
					if (!itLeft->second.IsCase() && itRight->second.IsCase()) {
						result = true;
					} else if (itLeft->second.IsCase() == itRight->second.IsCase()) {
						continue;
					} else {
						result = false;
					}
					break;
				}

				break;
			}

			if (sortKeyNum <= i) {
				result = true;
			}
		}

		if (!m_bAscending) {
			result = !result;
		}

		return result;
	}

private:
	// \[gL[
	ChkWordList::SortKey m_sortKey;
	// ǂ
	bool m_bAscending;
	// CHK [hꗗ
	std::map<std::string, ChkWord>& m_chkWordList;
};

//----------------------------------------------------------------------
// \[gs
//----------------------------------------------------------------------

void ChkWordList::Sort()
{
	CSLock cslock(m_pcsChkWordList);

	std::sort(m_sortedChkWord.begin(), m_sortedChkWord.end(), SortPredChkWord(m_sortKey, m_bAscending, m_chkWordList));
}

//----------------------------------------------------------------------
// \[gL[擾/ݒ
//----------------------------------------------------------------------

ChkWordList::SortKey ChkWordList::GetSortKey()
{
	CSLock cslock(m_pcsChkWordList);

	return m_sortKey;
}

void ChkWordList::SetSortKey(SortKey sortKey)
{
	CSLock cslock(m_pcsChkWordList);

	m_sortKey = sortKey;
}

//----------------------------------------------------------------------
// ǂ擾/ݒ
//----------------------------------------------------------------------

bool ChkWordList::IsAscending()
{
	CSLock cslock(m_pcsChkWordList);

	return m_bAscending;
}

void ChkWordList::SetAscending(bool bAscending)
{
	CSLock cslock(m_pcsChkWordList);

	m_bAscending = bAscending;
}

//----------------------------------------------------------------------
// n Ԗڂ CHK [hԂ
//----------------------------------------------------------------------

ChkWord& ChkWordList::GetAt(int index)
{
	CSLock cslock(m_pcsChkWordList);

	if (index < (int) m_sortedChkWord.size()) {
		std::map<std::string, ChkWord>::iterator it = m_chkWordList.find(m_sortedChkWord[index]);
		if (it != m_chkWordList.end()) {
			return it->second;
		}
	}

	// ւ͂Ȃ(͂)
	return m_emptyChkw;
}

//----------------------------------------------------------------------
// w CHK [h̃CfbNXԂ
//----------------------------------------------------------------------

int ChkWordList::GetIndex(const char* chkword)
{
	CSLock cslock(m_pcsChkWordList);

	std::vector<std::string>::iterator it = std::find(m_sortedChkWord.begin(), m_sortedChkWord.end(), chkword);
	if (it == m_sortedChkWord.end()) {
		return -1;
	}

	int index = std::distance(m_sortedChkWord.begin(), it);
	return index;
}

//----------------------------------------------------------------------
// n Ԗڂ CHK [h폜
//----------------------------------------------------------------------

bool ChkWordList::DeleteAt(int index)
{
	CSLock cslock(m_pcsChkWordList);

	if (index < (int) m_sortedChkWord.size()) {
		std::map<std::string, ChkWord>::iterator it = m_chkWordList.find(m_sortedChkWord[index]);
		if (it != m_chkWordList.end()) {
			std::string chkword = it->first;
			Delete(chkword.c_str());
			return true;
		}
	}

	return false;
}

